import torch
from torch import linalg as LA
import numpy as np

from omniisaacgymenvs.tasks.base.rl_task import RLTask
from ur5e_tool import UR5eTool as Robot

from omni.isaac.core.prims import RigidPrimView
from omni.isaac.core.articulations import ArticulationView
from omni.isaac.core.objects import DynamicCylinder, DynamicSphere
from omni.isaac.core.utils.prims import get_prim_at_path

from skrl.utils import omniverse_isaacgym_utils

# post_physics_step calls
# - get_observations()
# - get_states()
# - calculate_metrics()
# - is_done()
# - get_extras()


TASK_CFG = {"test": False,
            "device_id": 0,
            "headless": True,
            "sim_device": "gpu",
            "task": {"name": "MovingTarget",
                     "physics_engine": "physx",
                     "env": {"numEnvs": 1024,
                             "envSpacing": 1.5,
                             "episodeLength": 100,
                             "enableDebugVis": False,
                             "clipObservations": 1000.0,
                             "clipActions": 1.0,
                             "controlFrequencyInv": 4,
                             "actionScale": 2.5,
                             "dofVelocityScale": 0.1,
                             "controlSpace": "cartesian",
                            #  "goal": [0.5, 0.3, 0.0125]},
                             "goal": [0.5, 0.3, 0.0]},
                     "sim": {"dt": 0.0083,  # 1 / 120
                             "use_gpu_pipeline": True,
                             "gravity": [0.0, 0.0, -9.81],
                             "add_ground_plane": True,
                             "use_flatcache": True,
                             "enable_scene_query_support": False,
                             "enable_cameras": False,
                             "default_physics_material": {"static_friction": 1.0,
                                                         "dynamic_friction": 1.0,
                                                         "restitution": 0.0},
                             "physx": {"worker_thread_count": 4,
                                      "solver_type": 1,
                                      "use_gpu": True,
                                      "solver_position_iteration_count": 4,
                                      "solver_velocity_iteration_count": 1,
                                      "contact_offset": 0.005,
                                      "rest_offset": 0.0,
                                      "bounce_threshold_velocity": 0.2,
                                      "friction_offset_threshold": 0.04,
                                      "friction_correlation_distance": 0.025,
                                      "enable_sleeping": True,
                                      "enable_stabilization": True,
                                      "max_depenetration_velocity": 1000.0,
                                      "gpu_max_rigid_contact_count": 524288,
                                      "gpu_max_rigid_patch_count": 33554432,
                                      "gpu_found_lost_pairs_capacity": 524288,
                                      "gpu_found_lost_aggregate_pairs_capacity": 262144,
                                      "gpu_total_aggregate_pairs_capacity": 1048576,
                                      "gpu_max_soft_body_contacts": 1048576,
                                      "gpu_max_particle_contacts": 1048576,
                                      "gpu_heap_capacity": 33554432,
                                      "gpu_temp_buffer_capacity": 16777216,
                                      "gpu_max_num_partitions": 8},
                             "robot": {"override_usd_defaults": False,
                                       "fixed_base": False,
                                       "enable_self_collisions": True,
                                       "enable_gyroscopic_forces": True,
                                       "solver_position_iteration_count": 4,
                                       "solver_velocity_iteration_count": 1,
                                       "sleep_threshold": 0.005,
                                       "stabilization_threshold": 0.001,
                                       "density": -1,
                                       "max_depenetration_velocity": 1000.0,
                                       "contact_offset": 0.005,
                                       "rest_offset": 0.0},
                             "target": {"override_usd_defaults": False,
                                        "fixed_base": False,
                                        "enable_self_collisions": True,
                                        "enable_gyroscopic_forces": True,
                                        "solver_position_iteration_count": 4,
                                        "solver_velocity_iteration_count": 1,
                                        "sleep_threshold": 0.005,
                                        "stabilization_threshold": 0.001,
                                        "density": -1,
                                        "max_depenetration_velocity": 1000.0,
                                        "contact_offset": 0.005,
                                        "rest_offset": 0.0}}}}


class RobotView(ArticulationView):
    def __init__(self, prim_paths_expr: str, name: str = "robot_view") -> None:
        super().__init__(prim_paths_expr=prim_paths_expr, name=name, reset_xform_properties=False)


class MovingTargetTask(RLTask):
    def __init__(self, name, sim_config, env, offset=None) -> None:
        self._sim_config = sim_config
        self._cfg = sim_config.config
        self._task_cfg = sim_config.task_config

        self.dt = 1 / 120.0

        self._num_envs = self._task_cfg["env"]["numEnvs"]
        self._env_spacing = self._task_cfg["env"]["envSpacing"]
        self._action_scale = self._task_cfg["env"]["actionScale"]
        self._dof_vel_scale = self._task_cfg["env"]["dofVelocityScale"]
        self._max_episode_length = self._task_cfg["env"]["episodeLength"]
        self._control_space = self._task_cfg["env"]["controlSpace"]
        self._goal_position = self._task_cfg["env"]["goal"]

        # observation and action space
        self._num_observations = 19
        if self._control_space == "joint":
            self._num_actions = 6
        elif self._control_space == "cartesian":
            self._num_actions = 3
        else:
            raise ValueError("Invalid control space: {}".format(self._control_space))

        self._end_effector_link = "tool"

        RLTask.__init__(self, name, env)

    def set_up_scene(self, scene) -> None:
        self.get_robot()
        self.get_target()
        self.get_goal()

        # self.get_dummy()

        super().set_up_scene(scene)

        # robot view
        self._robots = RobotView(prim_paths_expr="/World/envs/.*/robot", name="robot_view")
        scene.add(self._robots)
        # end-effectors view
        self._end_effectors = RigidPrimView(prim_paths_expr="/World/envs/.*/robot/{}".format(self._end_effector_link), name="end_effector_view")
        scene.add(self._end_effectors)
        # hands view (cartesian)
        if self._control_space == "cartesian":
            self._tools = RigidPrimView(prim_paths_expr="/World/envs/.*/robot/tool", name="tool_view", reset_xform_properties=False)
            scene.add(self._tools)
        # target view
        self._targets = RigidPrimView(prim_paths_expr="/World/envs/.*/target", name="target_view", reset_xform_properties=False)
        scene.add(self._targets)
        # goal view
        self._goals = RigidPrimView(prim_paths_expr="/World/envs/.*/goal", name="goal_view", reset_xform_properties=False)
        scene.add(self._goals)

        # self._dummy = RigidPrimView(prim_paths_expr="/World/envs/.*/dummy", name="dummy_view", reset_xform_properties=False)
        # scene.add(self._dummy)


        self.init_data()

    def get_robot(self):
        robot = Robot(prim_path=self.default_zero_env_path + "/robot",
                      translation=torch.tensor([0.0, 0.0, 0.0]),
                      orientation=torch.tensor([1.0, 0.0, 0.0, 0.0]),
                      name="robot")
        self._sim_config.apply_articulation_settings("robot", get_prim_at_path(robot.prim_path), self._sim_config.parse_actor_config("robot"))

    def get_target(self):
        target = DynamicCylinder(
                    prim_path=self.default_zero_env_path + "/target",
                    # TODO: prim_path는 cloner 사용을 위해 바꿔야 함. 이하 참고 
                    name="target",
                    position=np.array([0.50, 0.20, 0.20]),
                    scale=np.array([0.03, 0.03, 0.03]),
                    color=np.array([255, 0, 0]),
                    density=100,
                    # mass=1.0,
                    )
        target.enable_rigid_body_physics()
        self._sim_config.apply_articulation_settings("target", get_prim_at_path(target.prim_path), self._sim_config.parse_actor_config("target"))
        # target.set_collision_enabled(True)
        # target.enable_rigid_body_physics()

    def get_dummy(self):
        dummy = DynamicCylinder(
                    prim_path=self.default_zero_env_path + "/dummy",
                    name="dummy",
                    position=np.array([0.80, 0.80, 0.20]),
                    scale=np.array([0.04, 0.04, 0.04]),
                    color=np.array([0, 0, 255]),
                    )
        dummy.enable_rigid_body_physics()
        self._sim_config.apply_articulation_settings("dummy", get_prim_at_path(dummy.prim_path), self._sim_config.parse_actor_config("dummy"))

    def get_goal(self):
        goal = DynamicSphere(prim_path=self.default_zero_env_path + "/goal",
                        name="goal",
                        radius=0.025,
                        color=torch.tensor([0, 255, 0]))
        self._sim_config.apply_articulation_settings("goal", get_prim_at_path(goal.prim_path), self._sim_config.parse_actor_config("goal"))
        # goal.set_collision_enabled(False)

    def init_data(self) -> None:
        self.robot_default_dof_pos = torch.tensor(np.radians([0.0, -100, 75, -90, -90, 0.0]), device=self._device, dtype=torch.float32)
        self.actions = torch.zeros((self._num_envs, self.num_actions), device=self._device)

        if self._control_space == "cartesian":
            self.jacobians = torch.zeros((self._num_envs, 10, 6, 6), device=self._device)
            self.hand_pos, self.hand_rot = torch.zeros((self._num_envs, 3), device=self._device), torch.zeros((self._num_envs, 4), device=self._device)

    def get_observations(self) -> dict:
        robot_dof_pos = self._robots.get_joint_positions(clone=False)
        robot_dof_vel = self._robots.get_joint_velocities(clone=False)
        end_effector_pos, end_effector_rot = self._end_effectors.get_world_poses(clone=False)
        target_pos, target_rot = self._targets.get_world_poses(clone=False)
        goal_pos, goal_rot = self._goals.get_world_poses(clone=False)

        dof_pos_scaled = 2.0 * (robot_dof_pos - self.robot_dof_lower_limits) \
            / (self.robot_dof_upper_limits - self.robot_dof_lower_limits) - 1.0
        dof_vel_scaled = robot_dof_vel * self._dof_vel_scale

        generalization_noise = torch.rand((dof_vel_scaled.shape[0], 6), device=self._device) + 0.5

        self.obs_buf[:, 0] = self.progress_buf / self._max_episode_length
        self.obs_buf[:, 1:7] = dof_pos_scaled[:, :6]
        self.obs_buf[:, 7:13] = dof_vel_scaled[:, :6] * generalization_noise
        self.obs_buf[:, 13:16] = target_pos - self._env_pos
        self.obs_buf[:, 16:19] = goal_pos - self._env_pos
        # self._env_pos is the position of the each environment. It comse from RLTask.

        # compute distance for calculate_metrics() and is_done()
        self._computed_tool_target_distance = LA.norm(end_effector_pos - target_pos, ord=2, dim=1)
        self._computed_target_goal_distance = LA.norm(target_pos - goal_pos, ord=2, dim=1)

        if self._control_space == "cartesian":
            self.jacobians = self._robots.get_jacobians(clone=False)
            self.hand_pos, self.hand_rot = self._tools.get_world_poses(clone=False)
            self.hand_pos -= self._env_pos

        return {self._robots.name: {"obs_buf": self.obs_buf}}

    def pre_physics_step(self, actions) -> None:
        reset_env_ids = self.reset_buf.nonzero(as_tuple=False).squeeze(-1)
        if len(reset_env_ids) > 0:
            self.reset_idx(reset_env_ids)

        self.actions = actions.clone().to(self._device)
        env_ids_int32 = torch.arange(self._robots.count, dtype=torch.int32, device=self._device)

        if self._control_space == "joint":
            targets = self.robot_dof_targets[:, :6] + self.robot_dof_speed_scales[:6] * self.dt * self.actions * self._action_scale

        elif self._control_space == "cartesian":
            goal_position = self.hand_pos + actions / 100.0
            delta_dof_pos = omniverse_isaacgym_utils.ik(jacobian_end_effector=self.jacobians[:, 10-1, :, :],  # tool index: 7
                                                        current_position=self.hand_pos,
                                                        current_orientation=self.hand_rot,
                                                        goal_position=goal_position,
                                                        goal_orientation=None)
            targets = self.robot_dof_targets[:, :6] + delta_dof_pos

        self.robot_dof_targets[:, :6] = torch.clamp(targets, self.robot_dof_lower_limits[:6], self.robot_dof_upper_limits[:6])
        # self.robot_dof_targets[:, 6:] = 0
        self._robots.set_joint_position_targets(self.robot_dof_targets, indices=env_ids_int32)
        # self._targets.enable_rigid_body_physics()
        # self._targets.enable_rigid_body_physics(indices=env_ids_int32)
        # self._targets.enable_gravities(indices=env_ids_int32)

    def reset_idx(self, env_ids) -> None:
        indices = env_ids.to(dtype=torch.int32)

        # reset robot
        pos = torch.clamp(self.robot_default_dof_pos.unsqueeze(0) + 0.25 * (torch.rand((len(env_ids), self.num_robot_dofs), device=self._device) - 0.5),
                          self.robot_dof_lower_limits, self.robot_dof_upper_limits)
        dof_pos = torch.zeros((len(indices), self._robots.num_dof), device=self._device)
        dof_pos[:, :] = pos
        dof_vel = torch.zeros((len(indices), self._robots.num_dof), device=self._device)
        self.robot_dof_targets[env_ids, :] = pos
        self.robot_dof_pos[env_ids, :] = pos

        self._robots.set_joint_position_targets(self.robot_dof_targets[env_ids], indices=indices)
        self._robots.set_joint_positions(dof_pos, indices=indices)
        self._robots.set_joint_velocities(dof_vel, indices=indices)

        # reset target
        target_pos = (torch.rand((len(env_ids), 3), device=self._device) - 0.5) * 2 \
                    * torch.tensor([0.25, 0.25, 0.10], device=self._device) \
                    + torch.tensor([0.50, 0.00, 0.20], device=self._device)

        self._targets.set_world_poses(target_pos + self._env_pos[env_ids], indices=indices)
        # self._targets.enable_rigid_body_physics()

        # reset goal
        goal_pos_xy_variation = torch.rand((len(env_ids), 2), device=self._device) * 0.1
        goal_pos_z_variation = torch.mul(torch.ones((len(env_ids), 1), device=self._device), 0.025)
        goal_pos_variation = torch.cat((goal_pos_xy_variation, goal_pos_z_variation), dim=1)
        goal_pos = torch.tensor(self._goal_position, device=self._device) + goal_pos_variation
        self._goals.set_world_poses(goal_pos + self._env_pos[env_ids], indices=indices)


        # reset dummy
        # self._dummy.set_world_poses()


        # bookkeeping
        self.reset_buf[env_ids] = 0
        self.progress_buf[env_ids] = 0

    def post_reset(self):
        self.num_robot_dofs = self._robots.num_dof
        self.robot_dof_pos = torch.zeros((self.num_envs, self.num_robot_dofs), device=self._device)
        dof_limits = self._robots.get_dof_limits()
        self.robot_dof_lower_limits = dof_limits[0, :, 0].to(device=self._device)
        self.robot_dof_upper_limits = dof_limits[0, :, 1].to(device=self._device)
        self.robot_dof_speed_scales = torch.ones_like(self.robot_dof_lower_limits)
        self.robot_dof_targets = torch.zeros((self._num_envs, self.num_robot_dofs), dtype=torch.float, device=self._device)
        # self._targets.enable_rigid_body_physics()
        # self._targets.enable_gravities()

        # randomize all envs
        indices = torch.arange(self._num_envs, dtype=torch.int64, device=self._device)
        self.reset_idx(indices)

    def calculate_metrics(self) -> None:
        self.rew_buf[:] = -self._computed_tool_target_distance - self._computed_target_goal_distance

    def is_done(self) -> None:
        self.reset_buf.fill_(0)
        # target reached
        self.reset_buf = torch.where(self._computed_target_goal_distance <= 0.035, torch.ones_like(self.reset_buf), self.reset_buf)
        # max episode length
        self.reset_buf = torch.where(self.progress_buf >= self._max_episode_length - 1, torch.ones_like(self.reset_buf), self.reset_buf)
